#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/ip_fw.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <net/if.h>
#include <linux/timer.h>
#include "csyncd.h"

extern int errno;
extern struct sockaddr_in sin_local, sin_remote, sin_reset;
extern int s;

#ifndef HAVE_STRUCT_IP_AUTOFW
struct ip_autofw {
	struct ip_autofw * next;
	u_int16_t type;
	u_int16_t low;
	u_int16_t hidden;
	u_int16_t high;
	u_int16_t visible;
	u_int16_t protocol;
	u_int32_t lastcontact;
	u_int32_t where;
	u_int16_t ctlproto;
	u_int16_t ctlport;
	u_int16_t flags;
	struct timer_list timer;
};

#define IP_FW_AUTOFW            5 
#define IP_AUTOFW_ADD           (IP_FW_APPEND | (IP_FW_AUTOFW << IP_FW_SHIFT))
#define IP_AUTOFW_DEL           (IP_FW_DELETE | (IP_FW_AUTOFW << IP_FW_SHIFT))
#define IP_AUTOFW_FLUSH         (IP_FW_FLUSH  | (IP_FW_AUTOFW << IP_FW_SHIFT))

#endif

/*
 * Translate a 32-bit inaddr to a dotted quad string
 */
char *intoa(int a)
{
    int b1,b2,b3,b4;
    static char buf[32];
    
    b1 = (a & 0xff000000) >> 24;
    b2 = (a & 0xff0000) >> 16;
    b3 = (a & 0xff00) >> 8;
    b4 = a & 0xff;
    sprintf(buf,"%d.%d.%d.%d",b1,b2,b3,b4);
    return buf;
}

/*
 * Send a positive response for datatype dt
 */
int send_ack(int dt)
{
    syncpkt_t pkt;
    int len;
    
    len = sizeof(pkt);
    pkt.datatype = dt;
    pkt.oper = OP_ACK;
    if(sendto(s,&pkt,len,0,(struct sockaddr *)&sin_remote,
      sizeof(sin_remote)) < 0) {
        return -1;
    }
    return 0;
}

/*
 * Send a negative response for datatype dt
 */
int send_nak(int dt)
{
    syncpkt_t pkt;
    int len;
    
    len = sizeof(pkt);
    pkt.datatype = dt;
    pkt.oper = OP_NAK;
    if(sendto(s,&pkt,len,0,(struct sockaddr *)&sin_remote,
      sizeof(sin_remote)) < 0) {
        return -1;
    }
    return 0;
}

/*
 * Receive waiting packet and decide what to do with it
 */
int receive_packet(void)
{
    int len,salen;
    syncpkt_t pkt;
    
    len=sizeof(pkt);
    salen=sizeof(sin_local);
    if(recvfrom(s,&pkt,len,0,(struct sockaddr *)&sin_local,&salen) < 0)
        return -1;
    switch(pkt.datatype) {
    case D_INVALID: /* ping from active, just ack it */
        send_ack(D_INVALID);
        break;
    case D_ARP: 
     {
        arp_data_t arp;
        memcpy(&arp,pkt.msg,sizeof(arp));
        arp_decode(pkt.oper,&arp);
        break;
     }
    case D_ROUTE: 
     {
        route_data_t rt;
        memcpy(&rt,pkt.msg,sizeof(rt));
        route_decode(pkt.oper,&rt);
        break;
     }
    case D_IPFWD: 
     {
        ipfwd_data_t ipf;
        memcpy(&ipf,pkt.msg,sizeof(ipf));
        ipfwd_decode(pkt.oper,&ipf);
        break;
     }
    case D_IPAUTOFW: 
     {
        ipautofw_data_t ipf;
        memcpy(&ipf,pkt.msg,sizeof(ipf));
        ipautofw_decode(pkt.oper,&ipf);
        break;
     }
    }
    return 0;
}

/*
 * Perform operation oper on the contents of an ARP packet
 */
void arp_decode(int oper,arp_data_t *arp)
{
    struct arpreq arp_req;
    int b1,b2,b3,b4,b5,b6;
    char ip[32],hw[32],mask[32];
    unsigned ipaddr,maskaddr;
    int arpsock;
    char buf[256];
    FILE *fp;
    
#ifdef DEBUG
    {
     char buf[128];
     sprintf(buf,"Received and processed ARP packet, op %d",oper);
     logstr(LOG_INFO,buf);
    }
#endif
    if(oper == OP_ADD || oper == OP_REMOVE) {
	memset(&arp_req,0,sizeof(arp_req));
	arp_req.arp_pa.sa_family = AF_INET;
	/* The ip has to start at sa_data[2] for some reason, and it
	   needs to be in network byte order */
	memcpy(&arp_req.arp_pa.sa_data[2],&arp->ipaddr,sizeof(arp->ipaddr));
	sscanf(arp->hwaddr,"%x:%x:%x:%x:%x:%x",&b1,&b2,&b3,&b4,&b5,&b6);
	arp_req.arp_ha.sa_family = ntohl(arp->hwtype);
	arp_req.arp_ha.sa_data[0] = b1;
	arp_req.arp_ha.sa_data[1] = b2;
	arp_req.arp_ha.sa_data[2] = b3;
	arp_req.arp_ha.sa_data[3] = b4;
	arp_req.arp_ha.sa_data[4] = b5;
	arp_req.arp_ha.sa_data[5] = b6;
	arp_req.arp_flags = ntohl(arp->flags);
	arp_req.arp_netmask.sa_family = AF_INET;
	/* the mask should follow the same rules as the IP */
	memcpy(&arp_req.arp_netmask.sa_data[2],&arp->mask,sizeof(arp->mask));
	strncpy(arp_req.arp_dev,arp->dev,sizeof(arp_req.arp_dev));
    }
    
    arpsock = socket(AF_INET,SOCK_DGRAM,0);
    if(arpsock < 0) {
        logstr(LOG_ERR,"Unable to open socket for ARP operations");
        send_nak(D_ARP);
        return;
    }

    switch(oper) {
    case OP_ADD: 
#ifdef DEBUG
	logstr(LOG_INFO,"Handling ARP ADD request");
#endif
	if(arp->ipaddr == (unsigned)(sin_reset.sin_addr.s_addr)) {
#ifdef DEBUG
	    logstr(LOG_INFO,"Rejecting own IP but ACKing anyway");
#endif
	    close(arpsock);
	    send_ack(D_ARP);
	    return;
	}
	if(ioctl(arpsock,SIOCSARP,&arp_req) < 0) {
	    sprintf(buf,"SIOCSARP: %s",sys_errlist[errno]);
	    logstr(LOG_INFO,buf);
	    close(arpsock);
	    send_nak(D_ARP);
	    return;
	}
	close(arpsock);
	send_ack(D_ARP); /* yay! */
	break;

    case OP_REMOVE:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling ARP REMOVE request");
#endif
	if(arp->ipaddr == (unsigned)(sin_reset.sin_addr.s_addr)) {
#ifdef DEBUG
	    logstr(LOG_INFO,"Rejecting own IP but ACKing anyway");
#endif
	    close(arpsock);
	    send_ack(D_ARP);
	    return;
	}
	if(ioctl(arpsock,SIOCDARP,&arp_req) < 0) {
	    sprintf(buf,"SIOCDARP: %s",sys_errlist[errno]);
	    logstr(LOG_INFO,buf);
	    close(arpsock);
	    send_nak(D_ARP);
	    return;
	}
	close(arpsock);
	send_ack(D_ARP); /* yay! */
	break;

    case OP_FLUSH:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling ARP FLUSH request");
#endif
        if((fp = fopen("/proc/net/arp","r")) == 0) {
            sprintf(buf,"/proc/net/arp: %s",sys_errlist[errno]);
            logstr(LOG_CRIT,buf);
            close(arpsock);
            send_nak(D_ARP);
            return;
        }

        if(fgets(buf,sizeof(buf),fp) != 0) {
            for(;fgets(buf,sizeof(buf),fp);) {
                memset(&arp_req,0,sizeof(arp_req));
                if(sscanf(buf,"%s 0x%x 0x%x %s %s %s\n", ip, (unsigned*)&arp_req.arp_ha.sa_family,
                 (unsigned*)&arp_req.arp_flags, hw, mask, arp_req.arp_dev) < 4)
                     break;
                 /* parse out ip addr */
                 sscanf(ip,"%d.%d.%d.%d",&b1,&b2,&b3,&b4);
                 ipaddr = htonl((b1<<24)+(b2<<16)+(b3<<8)+b4);
                 memcpy(&arp_req.arp_pa.sa_data[2],&ipaddr,sizeof(ipaddr));
                 arp_req.arp_pa.sa_family = AF_INET;
                 /* parse out hwaddr */
                 sscanf(hw,"%x:%x:%x:%x:%x:%x",&b1,&b2,&b3,&b4,&b5,&b6);
    	         arp_req.arp_ha.sa_data[0] = b1;
                 arp_req.arp_ha.sa_data[1] = b2;
                 arp_req.arp_ha.sa_data[2] = b3;
                 arp_req.arp_ha.sa_data[3] = b4;
                 arp_req.arp_ha.sa_data[4] = b5;
                 arp_req.arp_ha.sa_data[5] = b6;
                 arp_req.arp_netmask.sa_family = AF_INET;
                 if(mask[0] == '*')
                     memset(arp_req.arp_netmask.sa_data,0,sizeof(arp_req.arp_netmask.sa_data));
                 else {
                     sscanf(mask,"%d.%d.%d.%d",&b1,&b2,&b3,&b4);
                     maskaddr = htonl((b1<<24)+(b2<<16)+(b3<<8)+b4);
                     memcpy(&arp_req.arp_pa.sa_data[2],&maskaddr,sizeof(maskaddr));
                 }
                 if(ioctl(arpsock,SIOCDARP,&arp_req) < 0) {
                     sprintf(buf,"SIOCDARP: %s",sys_errlist[errno]);
                     logstr(LOG_CRIT,buf);
                     close(arpsock);
                     send_nak(D_ARP);
                     return;
                 }
#ifdef DEBUG                 
                 sprintf(buf,"Removed %s %s from ARP table",ip,hw);
                 logstr(LOG_INFO,buf);
#endif
            }
        }
        fclose(fp);
        close(arpsock);
	send_ack(D_ARP);
	break;
	
    default:
	logstr(LOG_ERR,"Unknown opcode");
	close(arpsock);
        send_nak(D_ARP);
        break;
    }
}

/*
 * Perform operation oper on the contents of a route packet
 */
void route_decode(int oper,route_data_t *rt)
{
    struct rtentry rte;
    int rtsock,ip,gw,mask,dummy;
    char buf[256],dev[32];
    struct ifreq ifr;
    unsigned my_addr, my_mask, my_net;
    FILE *fp;
    
#ifdef DEBUG
    sprintf(buf,"Received and processed ROUTE packet, op %d",oper);
    logstr(LOG_INFO,buf);
#endif
    if(oper == OP_ADD || oper == OP_REMOVE) {
	/* Set up the struct for ioctl().. refcnt & use are ignored */
	rte.rt_dev = rt->dev;
	rte.rt_dst.sa_family = AF_INET;
	memcpy(&rte.rt_dst.sa_data[2],&rt->dest,sizeof(rt->dest));
	rte.rt_gateway.sa_family = AF_INET;
	memcpy(&rte.rt_gateway.sa_data[2],&rt->gw,sizeof(rt->gw));
	rte.rt_flags = ntohl(rt->flags);
	rte.rt_metric = ntohl(rt->metric)+1;
	rte.rt_genmask.sa_family = AF_INET;
	memcpy(&rte.rt_genmask.sa_data[2],&rt->mask,sizeof(rt->mask));
	rte.rt_mtu = ntohl(rt->mtu);
	rte.rt_window = ntohl(rt->window);
	rte.rt_irtt = ntohl(rt->irtt);
    }

    /* Figure out the local net address so we don't delete it! :) */
    strcpy(ifr.ifr_name, "eth0");
    if((rtsock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
        logstr(LOG_ERR,"Unable to open socket for ROUTE operations");
        send_nak(D_ROUTE);
        return;
    }
    if(ioctl(rtsock,SIOCGIFADDR,&ifr) < 0) {
        sprintf(buf,"SIOCGIFADDR: %s",sys_errlist[errno]);
        logstr(LOG_ERR,buf);
        close(rtsock);
        send_nak(D_ROUTE);
        return;
    }
    memcpy(&my_addr,&ifr.ifr_addr.sa_data[2],sizeof(my_addr));
    if(ioctl(rtsock,SIOCGIFNETMASK,&ifr) < 0) {
        sprintf(buf,"SIOCGIFNETMASK: %s",sys_errlist[errno]);
        logstr(LOG_ERR,buf);
        close(rtsock);
        send_nak(D_ROUTE);
        return;
    }    
    memcpy(&my_mask,&ifr.ifr_netmask.sa_data[2],sizeof(my_mask));
    my_net = my_addr & my_mask;

    if(rt->dest == my_net) {
#ifdef DEBUG
	sprintf(buf,"Ignoring local network route %0x, ACKing anyway",ntohl(my_net));
	logstr(LOG_INFO,buf);
#endif
	send_ack(D_ROUTE);
	close(rtsock);
	return;
    }

    switch(oper) {
    case OP_ADD:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling ROUTE ADD request");
#endif
	if(ioctl(rtsock,SIOCADDRT,&rte) < 0) {
	    sprintf(buf,"SIOCADDRT: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(rtsock);
	    send_nak(D_ROUTE);
	    return;
	}
	close(rtsock);
    	send_ack(D_ROUTE);
    	break;
    	
    case OP_REMOVE:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling ROUTE REMOVE request");
#endif
	if(ioctl(rtsock,SIOCDELRT,&rte) < 0) {
	    sprintf(buf,"SIOCDELRT: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(rtsock);
	    send_nak(D_ROUTE);
	    return;
	}
	close(rtsock);
    	send_ack(D_ROUTE);
    	break;
    	
    case OP_FLUSH:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling ROUTE FLUSH request");
#endif
        if((fp = fopen("/proc/net/route","r")) == 0) {
            sprintf(buf,"/proc/net/route: %s",sys_errlist[errno]);
            logstr(LOG_CRIT,buf);
            close(rtsock);
            send_nak(D_ROUTE);
            return;
        }
	/* Parse the procfs data into a route entry struct */
        if(fgets(buf,sizeof(buf),fp) != 0) {
            for(;fgets(buf,sizeof(buf),fp);) {
                if(sscanf(buf,"%s %x %x %x %d %d %d %x %d %d %d\n", dev, 
                 &ip, &gw, (int*)&rte.rt_flags, &dummy, &dummy, (int*)&rte.rt_metric,
                 &mask, (int*)&rte.rt_mtu, (int*)&rte.rt_window, (int*)&rte.rt_irtt) < 10)
                    break;
                /* Ip (dest), gw & mask are already in net byte order */
                rte.rt_dev = dev;
                memcpy(&rte.rt_dst.sa_data[2],&ip,sizeof(ip));
                memcpy(&rte.rt_gateway.sa_data[2],&gw,sizeof(gw));
                memcpy(&rte.rt_genmask.sa_data[2],&mask,sizeof(mask));
                rte.rt_dst.sa_family = AF_INET;
                rte.rt_gateway.sa_family = AF_INET;
                rte.rt_genmask.sa_family = AF_INET;
                if(ip == my_net) {
#ifdef DEBUG
		    logstr(LOG_INFO,"Ignoring local network address");
#endif
		    continue;
		}
                if(ioctl(rtsock,SIOCDELRT,&rte) < 0) {
                    sprintf(buf,"SIOCDELRT: %s",sys_errlist[errno]);
                    logstr(LOG_CRIT,buf);
                    close(rtsock);
                    fclose(fp);
                    send_nak(D_ROUTE);
                    return;
                }
#ifdef DEBUG                 
                sprintf(buf,"Removed %s from ROUTE table",intoa(rt->dest));
                logstr(LOG_INFO,buf);
#endif
            }
        }
        fclose(fp);
        close(rtsock);
	send_ack(D_ROUTE);
	break;

    default:
	logstr(LOG_ERR,"Unknown opcode");
	close(rtsock);
    	send_nak(D_ROUTE);
    	break;
    }
}

/*
 * Perform operation oper on the contents of an ipfwd packet
 */
void ipfwd_decode(int oper,ipfwd_data_t *ipf)
{
    char buf[256];
    struct ip_fw ipfw;
    int sock,i;
    
#ifdef DEBUG
    sprintf(buf,"Received and processed IPFWD packet, op %d",oper);
    logstr(LOG_INFO,buf);
#endif

    if((sock=socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0) {
    	send_nak(D_IPFWD);
    	logstr(LOG_ERR,"Unable to open raw socket for IPFW operations");
    	return;
    }
    
    if(oper == OP_ADD || oper == OP_REMOVE) {
    	ipfw.fw_next = 0;
    	ipfw.fw_src.s_addr = ipf->saddr; /* s_addr is just an int */
    	ipfw.fw_dst.s_addr = ipf->daddr; /* we keep these in net order */
    	ipfw.fw_smsk.s_addr = ipf->smask;
    	ipfw.fw_dmsk.s_addr = ipf->dmask;
    	ipfw.fw_via.s_addr = ipf->via_addr;
    	ipfw.fw_viadev = 0; /* device pointer */
    	ipfw.fw_flg = ntohl(ipf->flags);
    	ipfw.fw_nsp = ntohl(ipf->nsp);
    	ipfw.fw_ndp = ntohl(ipf->ndp);
    	ipfw.fw_pcnt = ntohl(ipf->pcnt);
    	ipfw.fw_bcnt = ntohl(ipf->bcnt);
    	ipfw.fw_tosand = ntohl(ipf->tosand);
    	ipfw.fw_tosxor = ntohl(ipf->tosxor);
    	for(i=0;i<10;i++)
    	    ipfw.fw_pts[i] = ntohl(ipf->ports[i]);
    }

    switch(oper) {
    case OP_ADD:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling IPFWD ADD request");
#endif
	i = setsockopt(sock,IPPROTO_IP,IP_FW_APPEND_FWD,(char *)&ipfw,sizeof(ipfw));
	if(i != 0) {
	    sprintf(buf,"IPFWD ADD setsockopt: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(sock);
	    send_nak(D_IPFWD);
	    return;
	}
	close(sock);
	send_ack(D_IPFWD);
	break;
	
    case OP_REMOVE:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling IPFWD REMOVE request");
#endif
	i = setsockopt(sock,IPPROTO_IP,IP_FW_DELETE_FWD,(char *)&ipfw,sizeof(ipfw));
	if(i != 0) {
	    sprintf(buf,"IPFWD REMOVE setsockopt: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(sock);
	    send_nak(D_IPFWD);
	    return;
	}
	close(sock);
	send_ack(D_IPFWD);
	break;
    
    case OP_FLUSH:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling IPFWD FLUSH request");
#endif
	memset(&ipfw,0,sizeof(ipfw));
	i = setsockopt(sock,IPPROTO_IP,IP_FW_FLUSH_FWD,(char *)&ipfw,sizeof(ipfw));
	if(i != 0) {
	    sprintf(buf,"IPFWD FLUSH setsockopt: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(sock);
	    send_nak(D_IPFWD);
	    return;
	}
	close(sock);
	send_ack(D_IPFWD);
	break;

    default:
	logstr(LOG_ERR,"Unknown opcode");
    	send_nak(D_IPFWD);
    	close(sock);
    	break;
    }
}

/*
 * Perform operation oper on the contents of an ipautofw packet
 */
void ipautofw_decode(int oper,ipautofw_data_t *ipf)
{
    char buf[256];
    int i,sock;
    struct ip_autofw af;
    
#ifdef DEBUG
    sprintf(buf,"Received and processed IPAUTOFW packet, op %d",oper);
    logstr(LOG_INFO,buf);
#endif

    if((sock=socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0) {
    	send_nak(D_IPFWD);
    	logstr(LOG_ERR,"Unable to open raw socket for IPFW operations");
    	return;
    }
    
    if(oper == OP_ADD || oper == OP_REMOVE) {
    	af.type = ntohl(ipf->type);
    	af.low = ntohl(ipf->low);
    	af.hidden = ntohl(ipf->hidden);
    	af.high = ntohl(ipf->high);
    	af.visible = ntohl(ipf->visible);
    	af.protocol = ntohl(ipf->proto);
    	af.lastcontact = ntohl(ipf->lastcontact);
    	af.where = ntohl(ipf->where);
    	af.ctlproto = ntohl(ipf->ctlproto);
    	af.ctlport = ntohl(ipf->ctlport);
    	af.flags = ntohl(ipf->flags);
    	af.timer.next = 0;
    	af.timer.prev = 0;
    	af.timer.expires = ntohl(ipf->timer_exp);
    	af.timer.function = 0;
    }

    switch(oper) {
    case OP_ADD:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling IPAUTOFW ADD request");
#endif
	i = setsockopt(sock,IPPROTO_IP,IP_AUTOFW_ADD,(char *)&af,sizeof(af));
	if(i != 0) {
	    sprintf(buf,"IPAUTOFW ADD setsockopt: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(sock);
	    send_nak(D_IPAUTOFW);
	    return;
	}
	close(sock);
	send_ack(D_IPAUTOFW);
	break;
	
    case OP_REMOVE:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling IPAUTOFW REMOVE request");
#endif
	i = setsockopt(sock,IPPROTO_IP,IP_AUTOFW_DEL,(char *)&af,sizeof(af));
	if(i != 0) {
	    sprintf(buf,"IPAUTOFW REMOVE setsockopt: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(sock);
	    send_nak(D_IPAUTOFW);
	    return;
	}
	close(sock);
	send_ack(D_IPAUTOFW);
	break;
    
    case OP_FLUSH:
#ifdef DEBUG
	logstr(LOG_INFO,"Handling IPAUTOFW FLUSH request");
#endif
	memset(&af,0,sizeof(af));
	i = setsockopt(sock,IPPROTO_IP,IP_AUTOFW_FLUSH,(char *)&af,sizeof(af));
	if(i != 0) {
	    sprintf(buf,"IPAUTOFW FLUSH setsockopt: %s",sys_errlist[errno]);
	    logstr(LOG_ERR,buf);
	    close(sock);
	    send_nak(D_IPAUTOFW);
	    return;
	}
	close(sock);
	send_ack(D_IPAUTOFW);
	break;

    default:
	logstr(LOG_ERR,"Unknown opcode");
    	send_nak(D_IPAUTOFW);
    	close(sock);
    	break;
    }    
}
